"""Wrapper around boto3 to make it simpler to use."""

import boto3
import botocore
import io
import logging


from botocore.exceptions import ClientError
from dataclasses import dataclass, field
from typing import Generator, Optional


log = logging.getLogger(__name__)


@dataclass(frozen=True)
class BucketLocation:
    """Definition of a reference to an S3 bucket."""

    name: str = "s3os"
    region: Optional[str] = None


@dataclass(frozen=True)
class ObjectLocation:
    """Definition of a reference to a single S3 object."""

    key: str
    bucket: BucketLocation = field(default_factory=BucketLocation)


def create_bucket(bucket: BucketLocation) -> None:
    """
    Create an S3 bucket in a specified region.

    If a region is not specified, the bucket is created in the S3 default
    region (us-east-1).

    :param bucket: BucketLocation to create
    """
    try:
        if bucket.region is None:
            s3_client = boto3.client("s3")
            s3_client.create_bucket(Bucket=bucket.name)
        else:
            s3_client = boto3.client("s3", region_name=bucket.region)
            location = {"LocationConstraint": bucket.region}
            s3_client.create_bucket(
                Bucket=bucket.name, CreateBucketConfiguration=location
            )
    except ClientError as e:
        log.error(f"Failed to create bucket {bucket!r}. {e}")
        raise


def bucket_exists(bucket: BucketLocation) -> bool:
    """Return True if the given bucket exists. Otherwise False."""
    s3_client = boto3.client("s3")
    response = s3_client.list_buckets()

    return bucket.name in response["Buckets"]


def ensure_bucket(bucket: BucketLocation) -> None:
    """
    Ensure the given bucket exists.

    If the bucket already exists, then no action is taken.
    Otherwise, the bucket is created.
    """
    if not bucket_exists(bucket):
        create_bucket(bucket)


def upload_object(object_location: ObjectLocation, stream: io.BytesIO) -> None:
    """
    Upload the given data stream as an object to s3.

    :param object_location: Location of the object to create/update.
    :param stream: Byte steam of the object data.
    """
    s3 = boto3.client("s3")
    result = s3.upload_fileobj(stream, object_location.bucket.name, object_location.key)
    log.debug(f"Result of upload to {object_location}: {result}")


def download_object(object_location: ObjectLocation) -> io.BytesIO:
    """
    Download the given object from s3.

    Raises KeyError if the object could not be found.

    :param object_location: Location of the object to download.
    :return: Byte stream of the object data.
    """
    s3 = boto3.client("s3")
    stream = io.BytesIO()
    try:
        result = s3.download_fileobj(
            object_location.bucket.name, object_location.key, stream
        )
    except botocore.exceptions.ClientError as err:
        if err.response.get("Error", {}).get("Message", "Unknown") == "Not Found":
            log.debug(f"S3 object {object_location} does not exist.")
            raise KeyError(f"S3 object {object_location} does not exist.") from err
        raise

    stream.seek(0)
    log.debug(f"Result of download from {object_location}: {result}")
    return stream


def delete_object(object_location: ObjectLocation) -> None:
    """
    Download the given object from s3.

    :param object_location: Location of the object to delete.
    """
    s3 = boto3.client("s3")
    result = s3.delete_object(
        Bucket=object_location.bucket.name, Key=object_location.key
    )
    log.debug(f"Result of delete of {object_location}: {result}")


def generate_items_in_bucket(
    bucket: BucketLocation, prefix: Optional[str] = None
) -> Generator[ObjectLocation, None, None]:
    """
    Generate all object locations in a bucket.

    :param bucket: BucketLocation to inspect.
    :param prefix: Optional string prefix to filter the objects in the bucket by.
    :return: Generator of ObjectLocation for each object in the bucket.
    """
    s3 = boto3.client("s3")

    kwargs = {"Bucket": bucket.name}

    if prefix is not None:
        kwargs["Prefix"] = prefix

    while True:
        response = s3.list_objects(**kwargs)

        for obj in response["Contents"]:
            yield ObjectLocation(key=obj["Key"], bucket=bucket)

        # The S3 API is paginated, returning up to 1000 keys at a time.
        # Pass the continuation token into the next response, until we
        # reach the final page (when this field is missing).
        try:
            kwargs["ContinuationToken"] = response["NextContinuationToken"]
        except KeyError:
            break



"""Definition of a dict-like interface to s3."""

from collections import UserDict
from dataclasses import dataclass, field
from typing import Optional, Dict, Any
from uuid import uuid4

from s3os.api import store, retrieve, delete
from s3os.s3_wrapper import BucketLocation, ObjectLocation, generate_items_in_bucket


@dataclass
class S3DictConfig:
    """
    Configuration options for S3Dict.

    :param id: ID of the dict. Used to distinguish dicts within the same bucket.
        If not given, a universally unique identifier will be generated using UUID4.
        Objects in s3 can be retrieved by multiple S3Dict instances by using the same `id`.
    :param use_cache:
        Optionally data can be cached locally for improved speed. If `use_cache` is True then:

            - Set operations immediately upload to s3.
            - Get operations only download from s3 if they haven't been previously downloaded.
            - Delete operations immediately delete objects in s3.

        If `use_cache` is False then:

            - Set operations immediately upload to s3.
            - Get operations synchronously download objects from s3.
            - Delete operations immediately delete objects in s3.
    :param bucket: Optional. The s3 bucket to use.
    """

    id: str = field(default_factory=lambda: str(uuid4()))
    use_cache: bool = True
    bucket: BucketLocation = field(default_factory=BucketLocation)

    @property
    def s3_prefix(self):
        """The prefix to use for all items stored by this dict."""
        return f"{self.id}/"


class S3Dict(UserDict):
    """
    Provides a dict-like interface to objects stored in s3.

    Provides no guarantees or safety for concurrent access of objects.

    See S3DictConfig for configuration options.

    Deviations from the standard dict API:
      - del["item"] when "item" does not exist does not raise a KeyError.
      - __del__ doesn't delete all keys in s3 automatically. Use `clear()` to do this.
            This is different to normal behaviour where python garbage collection
            would delete all the items in a dictionary if they are no longer referenced.
    """

    def __init__(
        self, *args: Any, _config: Optional[S3DictConfig] = None, **kwargs: Any,
    ):
        """
        Create a new S3Dict.

        :param args: Standard `dict` arguments.
        :param _config: Optional S3DictConfig object. See S3DictConfig for default behaviour.
        :param kwargs: Standard `dict` keyword arguments.
        """
        if _config is None:
            _config = S3DictConfig()
        elif not isinstance(_config, S3DictConfig):
            raise ValueError(
                f"`_config` must be of type S3DictConfig or None. You passed: {_config=}."
            )
        self._config: S3DictConfig = _config

        # Call super after setting self._config so objects can be stored immediately.
        super(S3Dict, self).__init__(*args, **kwargs)

    def get_all_from_s3(self) -> Dict[str, Any]:
        """
        Discover all objects stored in s3 using this dict's ID.

        Caches the result on this object if configured to do so.

        Returns a dict of the discovered objects.
        """
        object_generator = generate_items_in_bucket(
            self._config.bucket, prefix=self._config.s3_prefix
        )

        all_items = dict()

        for object_location in object_generator:
            value = retrieve(object_location)
            key = self.convert_from_s3_key(object_location.key)
            all_items[key] = value

        if self._config.use_cache:
            # Update `data` directly rather than `self` so we don't just re-write
            # the keys back to s3 again.
            self.data.update(all_items)

        return all_items

    @property
    def as_dict(self) -> Dict[str, Any]:
        """
        Return a normal dict containing the information stored in s3 under this dict.

        Can be used on both cached and non-cached S3Dicts.

        NB: For cached dicts this will only return what is in the cache.
            Use `get_all_from_s3()` to force population of the entire dict.
        """
        if self._config.use_cache:
            return self.data
        else:
            return self.get_all_from_s3()

    def convert_from_s3_key(self, s3_key: str) -> str:
        """Strips the prepended dict ID from the s3 key."""
        if s3_key.startswith(self._config.s3_prefix):
            return s3_key.replace(self._config.s3_prefix, "", 1)
        return s3_key

    def convert_to_s3_key(self, key: str) -> str:
        """Prepend this dicts ID to the key for unique identification in s3."""
        return f"{self._config.s3_prefix}{key}"

    def __setitem__(self, key: str, value: Any) -> None:
        """Store the item in s3, as well as in the cache if configured to do so."""
        object_location = ObjectLocation(
            key=self.convert_to_s3_key(key), bucket=self._config.bucket
        )
        store(object_location, value)

        if self._config.use_cache:
            super(S3Dict, self).__setitem__(key, value)

    def __getitem__(self, item: str) -> Any:
        """Get the item from s3, using the cache if configured to do so."""
        object_location = ObjectLocation(
            key=self.convert_to_s3_key(item), bucket=self._config.bucket
        )
        if self._config.use_cache:
            try:
                # Try find it locally.
                value = super(S3Dict, self).__getitem__(item)
            except KeyError:
                # On failure, grab it from s3
                # If it doesn't exist in s3, then this will raise a KeyError itself
                # which is normal behaviour for a Dict.
                value = retrieve(object_location)
                # Store it locally to cache it for next time
                super(S3Dict, self).__setitem__(item, value)
        else:
            value = retrieve(object_location)

        return value

    def __delitem__(self, item: str) -> None:
        """Delete the item from s3, as well as from the cache if configured to do so."""
        object_location = ObjectLocation(
            key=self.convert_to_s3_key(item), bucket=self._config.bucket
        )
        # As s3 delete operations are a no-op if it doesn't exist, then we can't
        # tell if the item existed already or not.
        # Therefore this is a departure from the normal `dict` API because we can't
        # raise a KeyError on failure to delete.
        delete(object_location)
        if self._config.use_cache:
            try:
                super(S3Dict, self).__delitem__(item)
            except KeyError:
                # To keep things consistent, also don't raise a KeyError
                # when deleting from the cache.
                pass

    def clear(self) -> None:
        """Delete all the objects stored in s3 under this dict and clear the cache."""
        object_generator = generate_items_in_bucket(
            self._config.bucket, prefix=self._config.s3_prefix
        )

        for object_location in object_generator:
            delete(object_location)

        super(S3Dict, self).clear()



"""Python package to make storing and retrieving python objects in s3 trivial."""
# flake8: noqa

from .api import store, retrieve, delete, store_simple, retrieve_simple, delete_simple
from .s3_dict import S3Dict, S3DictConfig
from .s3_wrapper import BucketLocation, ObjectLocation



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['s3os']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.10.50,<2.0.0', 'ruamel.yaml>=0.16.5,<0.17.0', 'toml>=0.9,<0.10']

setup_kwargs = {
    'name': 's3os',
    'version': '1.0.0',
    'description': 'S3 Object Store. Store and retrieve python objects in s3 simply..',
    'long_description': 'S3 Object Store\n-------\n\n![badge](https://github.com/MartinHowarth/s3os/workflows/Test/badge.svg)\n\nSimple pythonic wrapper to use s3 as a python object store.\n\nProvides simple object storage methods; and a Dict-like interface for more complicated uses.\n\nExamples\n--------\nBasic usage:\n\n    from s3os import store_simple, retrieve_simple, delete_simple\n    \n    my_object = [1, 2, 3]\n    \n    store_simple("my_key", my_object)\n    \n    assert retrieve_simple("my_key") == my_object\n    \n    delete_simple("my_key")\n\nThe above example uses a global namespace in the bucket "s3os" - i.e. all the default settings of this package.\n\nYou can specify your own namespaces (i.e. buckets) as follows:\n\n    from s3os import store, retrieve, delete, ObjectLocation, Bucket\n    \n    my_bucket = Bucket("my_bucket")\n    my_object_location = ObjectLocation("my_key", bucket=my_bucket)\n    my_object = [1, 2, 3]\n    \n    store(my_object_location, my_object)\n    \n    assert retrieve(my_object_location) == my_object\n    \n    delete(my_object_location)\n\n\nOr simply use s3 like a normal python dictionary:\n\n    from s3os import S3Dict, S3DictConfig, Bucket\n    \n    my_bucket = Bucket("my_bucket")\n    s3dict = S3Dict(_config=S3DictConfig(id="my_dict_id", bucket=my_bucket))\n    \n    # Store information in s3\n    s3dict["apples"] = 5\n    s3dict["bananas"] = 2\n    \n    ...\n    \n    # Later, or in a different python executable, access the same dictionary again:\n    my_bucket = Bucket("my_bucket")\n    s3dict = S3Dict(_config=S3DictConfig(id="my_dict_id", bucket=my_bucket))\n    \n    print(s3dict["apples])  # 5\n    print(s3dict.get_all_from_s3())  # {"apples": 5, "bananas": 2}\n    \n\n\nBy default, `S3Dict` uses an internal cache to speed up item retrieval. \nSet and Delete operations are always performed synchronously.\n\n\nInstallation\n------------\nInstall the package:\n\n    pip install s3os\n\nSetup your AWS credentials. For example set these environment variables:\n\n    AWS_ACCESS_KEY_ID=...\n    AWS_SECRET_ACCESS_KEY=...\n    \nAlso see https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html for more authentication options.\n\n> Note: `boto.client()` and `Session` authentication methods are not currently supported - raise an issue or submit a PR if you want them!\n\n\nDevelopment installation\n------------------------\nInstall poetry - see https://pypi.org/project/poetry/\n\nThe following command should be used to install the dependencies:\n\n    poetry install\n\n\nTesting\n-------\nThe following command should be used to run the tests:\n\n    poetry run pytest tests\n\nValid AWS authentication credentials are required to run some of the tests.\nSee setup instructions.\n\nThe tests make a very small number of calls to S3, so the cost of running the tests is negligible.\n',
    'author': 'Martin Howarth',
    'author_email': 'howarth.martin@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MartinHowarth/s3os',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Definition of the simplest API to s3."""

from typing import Any

from .s3_wrapper import (
    ObjectLocation,
    ensure_bucket,
    upload_object,
    download_object,
    delete_object,
)
from .encoding import object_from_yaml_stream, object_to_yaml_stream


def store(object_location: ObjectLocation, obj: Any) -> None:
    """
    Store the given object in s3 at the given location.

    :param object_location: Definition of the bucket and key to store the object under.
    :param obj: The object to store. Must be able to be dumped/loaded to/from YAML.
    """
    ensure_bucket(object_location.bucket)
    obj_stream = object_to_yaml_stream(obj)
    upload_object(object_location, obj_stream)


def retrieve(object_location: ObjectLocation) -> Any:
    """
    Retrieve the object stored in s3 at the given location.

    :param object_location: Definition of the bucket and key to download.
    :return: The object retrieved, as a native python object.
    """
    obj_stream = download_object(object_location)
    obj = object_from_yaml_stream(obj_stream)
    return obj


def delete(object_location: ObjectLocation) -> None:
    """
    Delete the object stored in s3 at the given location.

    :param object_location: Definition of the bucket and key to delete.
    """
    delete_object(object_location)


def store_simple(key: str, value: Any) -> None:
    """
    Store the given object in s3 under the given key.

    Stores items at the root of the default bucket "s3os" with no unique identifier.

    :param key: Key to refer to the object.
    :param value: Object to store.
    """
    store(value, ObjectLocation(key=key))


def retrieve_simple(key: str) -> Any:
    """
    Store the given object in s3 under the given key.

    :param key: Key to of the object to retrieve
    """
    return retrieve(ObjectLocation(key=key))


def delete_simple(key: str) -> None:
    """
    Delete the object from s3 under the given key.

    :param key: Key to refer to the object.
    """
    delete(ObjectLocation(key=key))



"""Definition of dump and loading operations for storing objects in s3."""

import io

from ruamel import yaml

from typing import Any


def object_to_yaml_stream(obj: Any) -> io.BytesIO:
    """Convert the given object into a YAML byte stream."""
    stream = io.BytesIO()
    yaml.safe_dump(obj, stream=stream, encoding="utf-8")
    stream.seek(0)
    return stream


def object_from_yaml_stream(stream: io.BytesIO) -> Any:
    """Create an object from a stream of YAML bytes."""
    return yaml.safe_load(stream)
